from abc import ABCMeta
from functools import partial
from typing import Tuple, List, Callable, Optional

from .section import Section, _SECTION_RESULT
from .template import SectionTemplate
from ....utils import get_repr_info, FilePool, duplicates, wrap_empty


class _ISectionCollection(metaclass=ABCMeta):
    def __init__(self, items):
        """
        :param items: list of sections
        """
        self.__items = items

    def __repr__(self):
        """
        :return: get representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('sections', lambda: repr(tuple(section.name for section in self.__items))),
            ]
        )


_SECTION_GETTER = Callable[..., Section]
_SECTION_COLLECTION_RESULT = Tuple[bool, List[Tuple[str, _SECTION_RESULT]]]


class SectionCollectionTemplate(_ISectionCollection):
    def __init__(self, *items):
        """
        :param items: section templates
        """
        self.__items = [SectionTemplate.loads(item) for item in items]
        _ISectionCollection.__init__(self, self.__items)

    @property
    def items(self) -> List[SectionTemplate]:
        return list(self.__items)

    def __iter__(self):
        return self.items.__iter__()

    def __call__(self, scriptdir: str, identification=None, resources=None, environ=None,
                 **kwargs) -> 'SectionCollection':
        """
        generate section collection object
        :param scriptdir: script directory
        :param identification: identification
        :param resources: resource limits
        :param environ: environment variables
        :param kwargs: any other arguments
        :return: section collection object
        """
        if 'pool' in kwargs.keys():
            raise KeyError('Pool is not allowed to pass into section collection template.')

        _section_getters = [partial(
            item,
            scriptdir=scriptdir,
            identification=identification,
            resources=resources,
            environ=environ,
        ) for item in self.__items]

        _duplicated_names = duplicates([_getter().name for _getter in _section_getters])
        if _duplicated_names:
            raise KeyError('Duplicate names - {names}.'.format(names=repr(tuple(_duplicated_names))))

        return SectionCollection(*_section_getters)

    @classmethod
    def loads(cls, data) -> 'SectionCollectionTemplate':
        """
        load section collection template from data
        :param data: raw data
        :return: section collection template object
        """
        if isinstance(data, cls):
            return data
        elif isinstance(data, SectionTemplate):
            return cls(data)
        elif isinstance(data, (list, tuple)):
            return cls(*data)
        elif isinstance(data, dict):
            return cls(data)
        else:
            raise TypeError('Array or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


class SectionCollection(_ISectionCollection):
    def __init__(self, *getters: _SECTION_GETTER):
        """
        :param getters: functions to get section
        """
        self.__getters = list(getters)
        _ISectionCollection.__init__(self, [item() for item in self.__getters])

    @property
    def getters(self) -> List[_SECTION_GETTER]:
        return list(self.__getters)

    def __iter__(self):
        return self.getters.__iter__()

    def __call__(self, section_collection_start: Optional[Callable[['SectionCollection'], None]] = None,
                 section_collection_complete: Optional[
                     Callable[['SectionCollection', _SECTION_COLLECTION_RESULT], None]] = None,
                 **kwargs) -> _SECTION_COLLECTION_RESULT:
        """
        execute this section collection
        :return: success or not, full sections result
        """
        wrap_empty(section_collection_start)(self)
        with FilePool() as pool:
            _success = True
            _results = []
            for section_getter in self.__getters:
                section = section_getter(pool=pool)
                _section_success, _section_results, _section_info = section(**kwargs)
                _results.append((section.name, (_section_success, _section_results, _section_info)))
                if not _section_success:
                    _success = False
                    break

            _return = (_success, _results)

        wrap_empty(section_collection_start)(self, _return)
        return _return



from abc import ABCMeta
from typing import List, Optional, Callable

from .base import FileInputTemplate, FileInput
from .general import load_input_template
from ....utils import get_repr_info, wrap_empty


class _IFileInputCollection(metaclass=ABCMeta):
    def __init__(self, items):
        """
        :param items: file inputs
        """
        self.__items = items

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('inputs', lambda: len(self.__items)),
            ]
        )


class FileInputCollectionTemplate(_IFileInputCollection):
    def __init__(self, *items):
        """
        :param items: file input templates
        """
        self.__items = [load_input_template(item) for item in items]
        _IFileInputCollection.__init__(self, self.__items)

    @property
    def items(self) -> List[FileInputTemplate]:
        return list(self.__items)

    def __iter__(self):
        return self.items.__iter__()

    def __call__(self, **kwargs) -> 'FileInputCollection':
        """
        generate file input collection
        :param kwargs: plenty of arguments
        :return: file input collection
        """
        return FileInputCollection(*[item(**kwargs) for item in self.__items])

    @classmethod
    def loads(cls, data) -> 'FileInputCollectionTemplate':
        """
        load file input collection template from data
        :param data: raw data
        :return: file input collection template
        """
        data = data or []
        if isinstance(data, cls):
            return data
        elif isinstance(data, FileInputTemplate):
            return cls(data)
        elif isinstance(data, (list, tuple)):
            return cls(*data)
        elif isinstance(data, (dict, str)):
            return cls(load_input_template(data))
        else:
            raise TypeError('Array or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


class FileInputCollection(_IFileInputCollection):
    def __init__(self, *items):
        """
        :param items: file inputs
        """
        self.__items = items
        _IFileInputCollection.__init__(self, self.__items)

    @property
    def items(self) -> List[FileInput]:
        return list(self.__items)

    def __iter__(self):
        return self.items.__iter__()

    def __call__(self, input_collection_start: Optional[Callable[['FileInputCollection'], None]] = None,
                 input_collection_complete: Optional[Callable[['FileInputCollection'], None]] = None, **kwargs):
        """
        execute this file input setting
        """
        wrap_empty(input_collection_start)(self)
        for item in self.__items:
            item(**kwargs)
        wrap_empty(input_collection_complete)(self)



from abc import ABCMeta, abstractmethod


class SectionInfoTemplate(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, *args, **kwargs) -> 'SectionInfo':
        raise NotImplementedError


class SectionInfo(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, **kwargs):
        raise NotImplementedError



from abc import ABCMeta
from functools import partial
from typing import Mapping, Any

from ..control.model import RunResult
from ..service import DispatchTemplate, Command, CommandCollection, SectionInfoMapping, FileInput, \
    FileInputCollection, FileOutput, FileOutputCollection, Section, Task


class DispatchRunner(metaclass=ABCMeta):
    def __init__(self, template: DispatchTemplate):
        self.__template = template

    def _command_start(self, command: Command):
        pass

    def _command_complete(self, command: Command, result: RunResult):
        pass

    def _command_collection_start(self, collection: CommandCollection):
        pass

    def _command_collection_complete(self, collection: CommandCollection, result):
        pass

    def _info_mapping_start(self, mapping: SectionInfoMapping):
        pass

    def _info_mapping_complete(self, mapping: SectionInfoMapping, result):
        pass

    def _info_dump_start(self, section: Section, info: Mapping[str, Any]):
        pass

    def _info_dump_complete(self, section: Section, info: Mapping[str, Any]):
        pass

    def _input_start(self, input_: FileInput):
        pass

    def _input_complete(self, input_: FileInput):
        pass

    def _input_collection_start(self, collection: FileInputCollection):
        pass

    def _input_collection_complete(self, collection: FileInputCollection):
        pass

    def _output_start(self, output: FileOutput):
        pass

    def _output_complete(self, output: FileOutput):
        pass

    def _output_collection_start(self, collection: FileOutputCollection):
        pass

    def _output_collection_complete(self, collection: FileOutputCollection):
        pass

    def _section_start(self, section: Section):
        pass

    def _section_complete(self, section: Section, result):
        pass

    def _task_start(self, task: Task):
        pass

    def _task_complete(self, task: Task, result):
        pass

    def __call__(self, scriptdir: str, environ=None, environ_after=None, **kwargs):
        return partial(self.__template(
            scriptdir=scriptdir,
            environ=environ,
            environ_after=environ_after,
        ),
            command_start=self._command_start,
            command_complete=self._command_complete,
            command_collection_start=self._command_collection_start,
            command_collection_complete=self._command_collection_complete,
            info_mapping_start=self._info_mapping_start,
            info_mapping_complete=self._info_mapping_complete,
            info_dump_start=self._info_dump_start,
            info_dump_complete=self._info_dump_complete,
            input_start=self._input_start,
            input_complete=self._input_complete,
            input_collection_start=self._input_collection_start,
            input_collection_complete=self._input_collection_complete,
            output_start=self._output_start,
            output_complete=self._output_complete,
            output_collection_start=self._output_collection_start,
            output_collection_complete=self._output_collection_complete,
            section_start=self._section_start,
            section_complete=self._section_complete,
            task_start=self._task_start,
            task_complete=self._task_complete,
            **kwargs
        )



from abc import ABCMeta
from typing import Mapping, Any, Optional, Callable

from .base import SectionInfoTemplate, SectionInfo
from .general import load_info_template
from ....utils import get_repr_info, wrap_empty


class _ISectionInfoMapping(metaclass=ABCMeta):
    def __init__(self, items: Mapping[str, Any]):
        """
        :param items: mapping of section info objects
        """
        self.__items = items

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('keys', lambda: repr(tuple(sorted(self.__items.keys())))),
            ]
        )


class SectionInfoMappingTemplate(_ISectionInfoMapping):
    def __init__(self, **kwargs):
        """
        :param kwargs: mapping of section info template objects
        """
        self.__items = {key: load_info_template(data) for key, data in kwargs.items()}

        _ISectionInfoMapping.__init__(self, self.__items)

    @property
    def items(self) -> Mapping[str, SectionInfoTemplate]:
        return dict(self.__items)

    def __iter__(self):
        return self.items.items().__iter__()

    def __call__(self, **kwargs) -> 'SectionInfoMapping':
        """
        get section info info mapping object
        :param kwargs: plenty of arguments
        :return: section info info mapping object
        """
        return SectionInfoMapping(**{
            key: template(**kwargs) for key, template in self.__items.items()
        })

    @classmethod
    def loads(cls, data) -> 'SectionInfoMappingTemplate':
        """
        load section info mapping template from data
        :param data: raw data
        :return: section info mapping template
        """
        data = data or {}
        if isinstance(data, cls):
            return data
        elif isinstance(data, dict):
            return cls(**data)
        else:
            raise TypeError('Json or {type} expected but {actual} found.'.format(
                type=cls.__name__, actual=repr(type(data).__name__)))


_INFO_MAPPING_RESULT = Mapping[str, Any]


class SectionInfoMapping(_ISectionInfoMapping):
    def __init__(self, **kwargs):
        """
        :param kwargs: mapping of section info objects
        """
        self.__items = kwargs

        _ISectionInfoMapping.__init__(self, self.__items)

    @property
    def items(self) -> Mapping[str, SectionInfo]:
        return dict(self.__items)

    def __iter__(self):
        return self.items.items().__iter__()

    def __call__(self, info_mapping_start: Optional[Callable[['SectionInfoMapping'], None]] = None,
                 info_mapping_complete: Optional[Callable[['SectionInfoMapping', _INFO_MAPPING_RESULT], None]] = None,
                 **kwargs) -> _INFO_MAPPING_RESULT:
        """
        execute this info info
        """
        wrap_empty(info_mapping_start)(self)
        _ret = {key: info() for key, info in self.__items.items()}
        _result = {key: _data for key, (_ok, _data) in _ret.items() if _ok}

        wrap_empty(info_mapping_complete)(self, _result)
        return _result



import io

from .encoding import _try_read_to_bytes, _try_write
from ..model import RunResult
from ..process import common_process
from ...utils import eclosing


def common_run(args, shell: bool = False, stdin=None, stdout=None, stderr=None,
               environ=None, cwd=None, resources=None, identification=None) -> RunResult:
    """
    Create an common process with stream
    :param args: arguments for execution
    :param shell: use shell to execute args
    :param stdin: stdin stream (none means nothing)
    :param stdout: stdout stream (none means nothing)
    :param stderr: stderr stream (none means nothing)
    :param environ: environment variables
    :param cwd: new work dir
    :param resources: resource limit
    :param identification: user and group for execution
    :return: run result of this time
    """

    stdin_need_close = not stdin
    stdin = stdin or io.BytesIO()

    stdout_need_close = not stdout
    stdout = stdout or io.BytesIO()

    stderr_need_close = not stderr
    stderr = stderr or io.BytesIO()

    with eclosing(stdin, stdin_need_close) as stdin, \
            eclosing(stdout, stdout_need_close) as stdout, \
            eclosing(stderr, stderr_need_close) as stderr:
        with common_process(
                args=args, shell=shell,
                environ=environ, cwd=cwd,
                resources=resources, identification=identification,
        ) as cp:
            cp.communicate(_try_read_to_bytes(stdin), wait=False)
            cp.join()

            _try_write(stdout, cp.stdout)
            _try_write(stderr, cp.stderr)

        return cp.result



import os
from abc import ABCMeta
from typing import Optional, Mapping, Callable

from hbutils.string import env_template

from .base import FileInput, FileInputTemplate
from ...base import _check_workdir_file, _check_os_path, _process_environ
from ....utils import get_repr_info, makedirs, wrap_empty


class _ILinkFileInput(metaclass=ABCMeta):
    def __init__(self, file: str, local: str):
        """
        :param file: file path
        :param local: local path
        """
        self.__file = file
        self.__local = local

    def __repr__(self):
        """
        :return: representation string
        """
        return get_repr_info(
            cls=self.__class__,
            args=[
                ('file', lambda: repr(self.__file)),
                ('local', lambda: repr(self.__local)),
            ]
        )


class LinkFileInputTemplate(FileInputTemplate, _ILinkFileInput):
    def __init__(self, file: str, local: str):
        """
        :param file: file path
        :param local: local path
        """
        self.__file = file
        self.__local = local

        _ILinkFileInput.__init__(self, self.__file, self.__local)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def local(self) -> str:
        return self.__local

    # noinspection DuplicatedCode
    def __call__(self, scriptdir: str, workdir: str,
                 environ: Optional[Mapping[str, str]] = None, **kwargs) -> 'LinkFileInput':
        """
        generate link file input object from extension information
        :param scriptdir: script directory
        :param workdir: work directory
        :param environ: environment variable
        :return: link file input object
        """
        environ = _process_environ(environ)
        _file = os.path.normpath(
            os.path.abspath(os.path.join(scriptdir, _check_os_path(env_template(self.__file, environ)))))
        _local = os.path.normpath(
            os.path.abspath(os.path.join(workdir, _check_workdir_file(env_template(self.__local, environ)))))

        return LinkFileInput(
            file=_file, local=_local,
        )


class LinkFileInput(FileInput, _ILinkFileInput):
    def __init__(self, file: str, local: str, ):
        """
        :param file: file path
        :param local: local path
        """
        self.__file = file
        self.__local = local

        _ILinkFileInput.__init__(self, self.__file, self.__local)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def local(self) -> str:
        return self.__local

    def __call__(self, input_start: Optional[Callable[['LinkFileInput'], None]] = None,
                 input_complete: Optional[Callable[['LinkFileInput'], None]] = None, **kwargs):
        """
        execute this link event
        """
        wrap_empty(input_start)(self)
        _parent_path, _ = os.path.split(self.__local)
        makedirs(_parent_path)
        os.symlink(self.__file, self.__local, target_is_directory=os.path.isdir(self.__file))
        wrap_empty(input_complete)(self)



import os
from abc import ABCMeta
from typing import Optional, Mapping, Callable

from hbutils.string import env_template

from .base import FileOutputTemplate, FileOutput
from ...base import _check_workdir_file, _process_environ
from ....utils import get_repr_info, auto_copy_file, wrap_empty


def _check_os_path(path: str) -> str:
    """
    check file valid or not, when valid, just process it
    :param path: original file path
    :return: normalized file path
    """
    return os.path.normpath(path)


class _ICopyFileOutput(metaclass=ABCMeta):
    def __init__(self, local: str, file: str):
        """
        :param local: local path
        :param file: file path
        """
        self.__local = local
        self.__file = file

    def __repr__(self):
        """
        :return: representation string
        """

        return get_repr_info(
            cls=self.__class__,
            args=[
                ('local', lambda: repr(self.__local)),
                ('file', lambda: repr(self.__file)),
            ]
        )


class CopyFileOutputTemplate(FileOutputTemplate, _ICopyFileOutput):
    def __init__(self, local: str, file: str):
        """
        :param local: local path
        :param file: file path
        """
        self.__local = local
        self.__file = file

        _ICopyFileOutput.__init__(self, self.__local, self.__file)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def local(self) -> str:
        return self.__local

    # noinspection DuplicatedCode
    def __call__(self, scriptdir: str, workdir: str, environ: Optional[Mapping[str, str]] = None,
                 **kwargs) -> 'CopyFileOutput':
        """
        generate copy file output object from extension information
        :param scriptdir: script directory
        :param workdir: work directory
        :param environ: environment variable
        :return: copy file output object
        """
        environ = _process_environ(environ)
        _local = os.path.normpath(
            os.path.abspath(os.path.join(workdir, _check_workdir_file(env_template(self.__local, environ)))))
        _file = os.path.normpath(
            os.path.abspath(os.path.join(scriptdir, _check_os_path(env_template(self.__file, environ)))))

        return CopyFileOutput(file=_file, local=_local)


class CopyFileOutput(FileOutput, _ICopyFileOutput):
    def __init__(self, local: str, file: str):
        """
        :param local: local path
        :param file: file path
        """
        self.__local = local
        self.__file = file

        _ICopyFileOutput.__init__(self, self.__local, self.__file)

    @property
    def file(self) -> str:
        return self.__file

    @property
    def local(self) -> str:
        return self.__local

    def __call__(self, output_start: Optional[Callable[['CopyFileOutput'], None]] = None,
                 output_complete: Optional[Callable[['CopyFileOutput'], None]] = None, **kwargs):
        """
        execute this file output
        """
        wrap_empty(output_start)(self)
        auto_copy_file(self.__local, self.__file)
        wrap_empty(output_complete)(self)



import codecs
import json
import os
from typing import List, Optional

import click
from click.core import Context, Option

from .environ import _load_environ
from .event import _load_dispatch_getter
from .exception import _raise_exception_with_exit_code
from ..event import _DEFAULT_FILENAME
from ..script.result import result_to_json
from ...config.meta import __TITLE__, __VERSION__, __AUTHOR__, __AUTHOR_EMAIL__


# noinspection DuplicatedCode,PyUnusedLocal
def print_version(ctx: Context, param: Option, value: bool) -> None:
    """
    Print version information of cli
    :param ctx: click context
    :param param: current parameter's metadata
    :param value: value of current parameter
    """
    if not value or ctx.resilient_parsing:
        return
    click.echo('{title}, version {version}.'.format(title=__TITLE__.capitalize(), version=__VERSION__))
    click.echo('Developed by {author}, {email}.'.format(author=__AUTHOR__, email=__AUTHOR_EMAIL__))
    ctx.exit()


CONTEXT_SETTINGS = dict(
    help_option_names=['-h', '--help']
)

_DEFAULT_TASK = 'main'


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('-v', '--version', is_flag=True,
              callback=print_version, expose_value=False, is_eager=True,
              help="Show package's version information.")
@click.option('-s', '--script', type=click.Path(exists=True, readable=True),
              help='Path of pji script.', default=_DEFAULT_FILENAME, show_default=True)
@click.option('-t', '--task', type=str, help='Task going to be executed.',
              default=_DEFAULT_TASK, show_default=True)
@click.option('-e', '--environ', type=str, multiple=True,
              help='Environment variables (loaded before global config).')
@click.option('-E', '--environ_after', type=str, multiple=True,
              help='Environment variables (loaded after global config).')
@click.option('-i', '--information', type=click.Path(dir_okay=False),
              help='Information dump file (no dumping when not given).')
def cli(script: str, task: str, environ: List[str], environ_after: List[str],
        information: Optional[str] = None):
    _dispatch_getter = _load_dispatch_getter(script)
    _success, _result = _dispatch_getter(
        environ=_load_environ(environ),
        environ_after=_load_environ(environ_after),
    )(task)

    if information:
        click.echo(
            click.style('Dumping result of this work to {info} ... '.format(
                info=repr(os.path.abspath(information))), bold=False),
            nl=False,
        )

        with codecs.open(information, 'w') as info_file:
            json.dump(result_to_json(_success, _result), info_file, indent=4, sort_keys=True)

        click.echo(click.style('COMPLETE', fg='green'), nl=True)

    if _success:
        click.echo(click.style('Task success.', fg='green'))
    else:
        click.echo(click.style('Task failed.', fg='red'))
        raise _raise_exception_with_exit_code(1, 'task failed.')



from typing import Type, Optional, Callable

from ..event import _load_dispatch_getter as _load_dispatch_abstract_getter
from ..runner import DispatchRunner
from ...service import Dispatch


class _DefaultDispatchEventRunner(DispatchRunner):
    pass


def _load_dispatch_getter(filename: str = None, runner_class: Optional[Type[DispatchRunner]] = None) \
        -> Callable[..., Dispatch]:
    return _load_dispatch_abstract_getter(runner_class or _DefaultDispatchEventRunner, filename)



from abc import ABCMeta, abstractmethod


class FileOutputTemplate(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, *args, **kwargs) -> 'FileOutput':
        raise NotImplementedError


class FileOutput(metaclass=ABCMeta):
    @abstractmethod
    def __call__(self, **kwargs):
        raise NotImplementedError
